%==========================================================================
% StrIkE-GOLDD (Structural Identifiability taken as Extended-Generalized
% Observability with Lie Derivatives and Decomposition)
%--------------------------------------------------------------------------
%
% A Matlab toolbox for structural identifiability analysis of nonlinear models
%
%--------------------------------------------------------------------------
%
% User input can be specified in file '/functions/options.m'
%
%--------------------------------------------------------------------------
% Version 0.1. 
% Last modified: 16/05/2016
% Alejandro Fernandez Villaverde (afernandez@uvigo.es)
%==========================================================================

function STRIKE_GOLDD

fprintf('\n\n ------------------------- \n');
fprintf(' >>> StrIkE-GOLDD TOOLBOX \n');
fprintf(' ------------------------- \n');

%==========================================================================
% Clear workspace, read options, add folders to path:
clear
tStart = tic;
global x f p maxstates
options;
maxstates = opts.maxstates;
addpath(genpath(paths.meigo));
addpath(genpath(paths.models));
addpath(genpath(paths.results));
addpath(genpath(paths.functions));

%==========================================================================
% Load model:
load(model)
fprintf('\n Analyzing identifiability of %s ... \n', model);

tic
%==========================================================================
% Initialize variables:
identifiables  = [];       % identifiable parameters.
nonidentif     = [];       % unidentifiable parameters.
obs_states     = [];       % observable states.
unobs_states   = [];       % unobservable states.
lastrank       = NaN;
decomp_flag    = 0;
unidflag       = 0;
    
% ==========================================================================
% Remove parameters that have already been classified as identifiable:
if exist('prev_ident_pars','var') == 1
    for np=1:numel(prev_ident_pars)
        [~, original_index] = ismember(prev_ident_pars(np),p); 
        p(original_index)=[];                
    end
end
    
% ==========================================================================
% Remove states that have already been classified as observable:
if exist('prev_obs_states','var') == 1
    for nx=1:numel(prev_obs_states)
        [~, original_index] = ismember(prev_obs_states(nx),x); 
        x(original_index)=[];          
        f(original_index)=[];  
    end
end

%==========================================================================
% Dimensions of the problem: 
m = numel(h);                     % number of outputs
n = numel(x);                     % number of states
q = numel(p);                     % number of unknown parameters
r = n+q;                          % number of states + parameters
xaug = [x;p];                     % states and parameters 
faug = [f; zeros(numel(p),1)];    % augmented dynamics

fprintf('\n >>> The model contains:\n %d states:\n %s',numel(x),char(x));
fprintf('\n %d outputs:\n %s',numel(h),char(h));
fprintf('\n %d parameters:\n %s',q,char(p));

total_expr = [f; h];
variaveis = symvar(total_expr).'; % All variables appearing in the model
                                  % (may include e.g. inputs or parameters
                                  % not considered for identification, etc)

%==========================================================================
% Check which states are directly measured, if any: 
saidas     = cell(m,1);
ismeasured = zeros(1,n);
for i=1:m
    saidas{i} = char(h(i));
end
for i=1:n
    ismeasured(i) = sum(strcmpi(char(x(i)),saidas));
end
meas_x_indices = find(ismeasured);       % indices of the measured states
unmeas_x_indices = find((1-ismeasured)); % indices of the unmeasured states
meas_x = x(meas_x_indices);              % names of the measured states

%==========================================================================
if opts.forcedecomp == 0 

    %==========================================================================
    % Build Identifiability-Observability matrix, Oi(x)
    nd = ceil(r/m-1);  % minimum number of Lie derivatives
    increaseLie = 1;
    fprintf('\n\n >>> Building the observability-identifiability matrix requires at least %d Lie derivatives',nd);
    fprintf('\n     Calculating derivatives: ');
    tic
    onx = zeros(m*(1+nd),r);
    onx = sym(onx);
    onx(1:m,:) = jacobian(h,xaug); % 1st block
    totaltime = toc;
    ind = 2;
    lasttime = 0;
    while ind < (nd+2) && lasttime < opts.maxLietime % 2nd and subsequent blocks
        tic
        Lieh = onx(((ind-2)*m+1):(ind-1)*m,:)*faug;
        onx(((ind-1)*m+1):(ind*m),:) = jacobian(Lieh,xaug);
        lasttime  = toc;
        totaltime = totaltime + lasttime;
        ind = ind+1;
        fprintf('%d ',ind-2); 
    end
    if ind == (nd+2)
        obsidentmatrix = sprintf('obs_ident_matrix_%s_%d_Lie_deriv',model,nd);
        obsidentfile = strcat(pwd,'\results\',obsidentmatrix);
        save(obsidentfile);
        while increaseLie == 1
            fprintf('\n >>> Observability-Identifiability matrix built with %d Lie derivatives',nd);
            fprintf('\n     (calculated in %d seconds)',totaltime);
            
            %==========================================================================
            % Check identifiability by calculating rank:
            fprintf('\n >>> Calculating rank...');
            tic
            if opts.knowninitc == 1
                xind = find(known_ics);
                if size(ics) ~= size(x)
                    ics = transpose(ics);
                end
                onx = subs(onx,x(xind),ics(xind));
            end  
            if opts.numeric == 1
                load primes10k
                primes = primos(1:numel(variaveis));
                numonx = subs(onx,variaveis,primes);  
            else
                numonx = onx;
            end
            rango = double(rank(numonx));
            fprintf('\n     Rank = %d (calculated in %d seconds)',rango,toc);         
            if rango == r 
                obs_states = x;
                identifiables = p;
                increaseLie = 0;
            else% With that number of Lie derivatives the array is not full rank.                       
                % If possible, calculate one more Lie derivative and retry:
                if nd < r  && lasttime < opts.maxLietime && rango ~= lastrank
                    tic
                    nd     = nd + 1;
                    i      = 1+nd;
                    newLie = onx(((i-2)*m+1):(i-1)*m,:)*faug;
                    newOnx = jacobian(newLie,xaug);
                    onx    = [onx; newOnx];   
                    clear newLie newOnx
                    lasttime  = toc;
                    totaltime = totaltime + lasttime;
                    obsidentmatrix = sprintf('obs_ident_matrix_%s_%d_Lie_deriv',model,nd);
                    obsidentfile = strcat(pwd,'\results\',obsidentmatrix);
                    save(obsidentfile);
                    lastrank = rango;
                % If that is not possible, there are several possible causes:
                else 
                    if nd >= r % The maximum number of Lie derivatives has been reached
                        unidflag = 1; 
                        fprintf('\n    The model is structurally unidentifiable as a whole');
                    else
                        if rango == lastrank
                            numonx = numonx(1:(end-m),:);
                            nd = nd - 1;
                            unidflag = 1; % note that the pars may still be identifiable (rank deficiency may be due to initial conditions)
                        else
                            if lasttime >= opts.maxLietime
                                fprintf('\n    => More Lie derivatives would be needed to see if the model is structurally unidentifiable as a whole.');
                                fprintf('\n    However, the maximum computation time allowed for calculating each of them has been reached.');
                                fprintf('\n    You can increase it by changing <<opts.maxLietime>> (currently opts.maxLietime = %d)',opts.maxLietime);
                            end
                        end
                    end                    
                    % Eliminate columns one by one to check identifiability of the associated parameters: 
                    [identifiables,nonidentif,obs_states,unobs_states] = ...
                         elim_and_recalc(unmeas_x_indices,rango,numonx,opts);
                     increaseLie = 0;
                end
            end
        end        
    else
        if opts.decomp == 0
            fprintf('\n >>> Calculating rank...');
            tic
            if opts.knowninitc == 1
                xind = find(known_ics);
                if size(ics) ~= size(x)
                    ics = transpose(ics);
                end
                onx = subs(onx,x(xind),ics(xind));
            end  
            if opts.numeric == 1
                load primes10k
                primes = primos(1:numel(variaveis));
                numonx = subs(onx,variaveis,primes);  
            else
                numonx = onx;
            end
            rango = double(rank(numonx));
            fprintf('\n     Rank = %d (calculated in %d seconds)',rango,toc); 
            [identifiables,nonidentif,obs_states,unobs_states] = ...
            elim_and_recalc(unmeas_x_indices,rango,numonx,opts);
        else
            decomp_flag = 1; 
        end
    end
end % end of 'if opts.forcedecomp == 0'

if decomp_flag == 1 || opts.forcedecomp == 1 
    fprintf('\n The model will be decomposed. \n');
    if opts.decomp_user == 1
        model_results_folder = sprintf('results/decomp_user_%s_%s_maxLietime_%d',model,date,opts.maxLietime);
        if exist('submodels','var') == 0 || numel(submodels) == 0
            warning('You have selected to enter the submodels manually. However, no submodels have been specified => STRIKE-GOLDD will decompose the model using optimization. If you want to avoid it, create a non-empty submodels vector in the options file.') 
            opts.decomp_user = 0;
        end
    end
    if opts.decomp_user == 0
        model_results_folder = sprintf('results/decomp_%s_%s_maxstates_%d_maxLietime_%d',model,date,maxstates,opts.maxLietime);
        submodels = [];
    end
    mkdir(model_results_folder);
    [identifiables,nonidentif,obs_states,unobs_states] = decomp(model,opts,model_results_folder,submodels);
end

%==========================================================================
% Build the vectors of identifiable / non identifiable parameters and
% initial conditions (may be repeated if decomposition was used):
p_id         = symvar(identifiables);        
p_un         = symvar(nonidentif);          
obs_states   = symvar(obs_states);       
unobs_states = symvar(unobs_states); 

%==========================================================================
% Report results:
fprintf('\n\n ------------------------ \n');
fprintf(' >>> RESULTS SUMMARY:\n');
fprintf(' ------------------------ \n');
load(model)

if numel(p_id) == numel(p)
    fprintf('\n >>> The model is structurally identifiable:');
    fprintf('\n     All its parameters are structurally identifiable. \n');
else    
    if unidflag == 1 
        fprintf('\n >>> The model is structurally unidentifiable.');
        fprintf('\n >>> These parameters are identifiable:\n      %s ',char(p_id));
        fprintf('\n >>> These parameters are unidentifiable:\n      %s \n',char(p_un));
    else             
        fprintf('\n >>> These parameters are identifiable:\n      %s ',char(p_id));
    end  
    
if numel(obs_states)>0,    fprintf('\n >>> These states are observable (and their initial conditions, if considered unknown, are identifiable):\n      %s ',char(obs_states)); end
if numel(unobs_states)>0,  fprintf('\n >>> These states are unobservable (and their initial conditions, if considered unknown, are unidentifiable):\n      %s ',char(unobs_states)); end
if numel(meas_x)>0,        fprintf('\n >>> These states are directly measured:\n      %s ',char(meas_x)); end
          
%==========================================================================
% Search for identifiable parameter combinations (combos):
    if opts.findcombos == 1 && exist('onx','var') == 1
        % Save results first, just in case the user kills the process:
        resultsname = sprintf('id_results_%s',model);
        fullresultsname = strcat(pwd,'\results\',resultsname);
        save(fullresultsname)      
        [parpde,stringpde] = combos(p_un,onx,n);
        if numel(parpde) == 0
            fprintf('\n\n >>> No identifiable combinations of parameters could be found');
        else
            fprintf('\n\n >>> There are identifiable combinations of parameters.');
            fprintf('\n     They can be found by solving the following PDE(s):');
            for numpdes = 1:numel(stringpde)
                fprintf('\n     %s \n',char(stringpde));
            end
        end       
    end
end

%==========================================================================
totaltime = toc(tStart);
fprintf('\n Total execution time: %d \n\n',totaltime);

%==========================================================================
% Save results:
if decomp_flag == 0
    resultsname = sprintf('id_results_%s_%s',model,date);
else
    resultsname = sprintf('id_results_%s_maxstates_%d_%s',model,maxstates,date);
    delete VNS_report.mat
end    
fullresultsname = strcat(pwd,'\results\',resultsname);
save(fullresultsname);

end    
